@@ -65,8 +65,11 @@ bool SetBG::set_bg(	Glib::ustring &disp, Glib::ustring file, SetMode mode, Gdk::
 	window->get_geometry(winx,winy,winw,winh,wind);
 
 	// check to see if nautilus is running
-	if (SetBG::nautilus_running(window))
-		return SetBG::set_bg_nautilus(screen, file, mode, bgcolor);		
+	if (SetBG::get_rootwindowtype(window) == SetBG::NAUTILUS)
+		return SetBG::set_bg_nautilus(screen, file, mode, bgcolor);
+
+    // NOTE: although we now have a way of detecting XFCE, just setting the bg like
+    // in normal X seems to work just fine.
 
 	// create gc and colormap
 	gc_ = Gdk::GC::create(window);
@@ -114,6 +117,10 @@ bool SetBG::set_bg(	Glib::ustring &disp, Glib::ustring file, SetMode mode, Gdk::
 		case SetBG::SET_ZOOM:
 			outpixbuf = SetBG::make_zoom(pixbuf, winw, winh, bgcolor);
 			break;
+
+		case SetBG::SET_ZOOM_FILL:
+			outpixbuf = SetBG::make_zoom_fill(pixbuf, winw, winh, bgcolor);
+			break;
 			
 		default:
 			std::cerr << _("Unknown mode for saved bg on") << " " << disp << std::endl;
@@ -334,6 +341,10 @@ bool SetBG::set_bg_xinerama(XineramaScreenInfo* xinerama_info, gint xinerama_num
 		case SetBG::SET_ZOOM:
 			outpixbuf = SetBG::make_zoom(pixbuf, tarw, tarh, bgcolor);
 			break;
+
+		case SetBG::SET_ZOOM_FILL:
+			outpixbuf = SetBG::make_zoom_fill(pixbuf, tarw, tarh, bgcolor);
+			break;
 			
 		default:
 			std::cerr << _("Unknown mode for saved bg") << " " << std::endl;
@@ -373,85 +384,109 @@ bool SetBG::set_bg_xinerama(XineramaScreenInfo* xinerama_info, gint xinerama_num
  *
  * @returns 	True if nautilus is drawing the desktop.
  */
-bool SetBG::nautilus_running(Glib::RefPtr<Gdk::Window> rootwin)
-{   
+SetBG::RootWindowType SetBG::get_rootwindowtype(Glib::RefPtr<Gdk::Window> rootwin)
+{
 	GdkAtom type;
     gint format;
     gint length;
 	guchar *data;
-    gboolean retval;
+    SetBG::RootWindowType retval = SetBG::DEFAULT;
+    gboolean ret = FALSE;
 
-	retval = gdk_property_get(rootwin->gobj(), 
+	ret =    gdk_property_get(rootwin->gobj(),
 						      gdk_atom_intern("NAUTILUS_DESKTOP_WINDOW_ID", FALSE),
                               gdk_atom_intern("WINDOW", FALSE),
 							  0,
-							  4, /* wtf is a length of a window */
+							  4, /* length of a window is 32bits*/
 							  FALSE, &type, &format, &length, &data);
 
-    if (retval)
-        g_free(data);
+    if (!ret)
+        return SetBG::DEFAULT;
 
+    guint wid = *(guint*)data;
+
+    Display *xdisp = GDK_DISPLAY_XDISPLAY(rootwin->get_display()->gobj());
+    Atom propatom = XInternAtom(xdisp, "WM_CLASS", FALSE);
+
+    XTextProperty tprop;
+
+    gchar **list;
+    gint num;
+
+    if (XGetTextProperty(xdisp, wid, &tprop, propatom) && tprop.nitems)
+    {
+        if (XTextPropertyToStringList(&tprop, &list, &num))
+        {
+            // expect 2 strings here (XLib tells us there are 3)
+            if (num != 3)
+                retval = SetBG::DEFAULT;
+            else
+            {
+                std::string strclass = std::string(list[1]);
+                if (strclass == std::string("Xfdesktop")) retval = SetBG::XFCE;     else
+                if (strclass == std::string("Nautilus"))  retval = SetBG::NAUTILUS; else
+                {
+                    std::cerr << _("UNKNOWN ROOT WINDOW TYPE DETECTED, will attempt to set via normal X procedure") << "\n";
+                    retval = SetBG::UNKNOWN;
+                }
+
+         }
+
+            XFreeStringList(list);
+        }
+        XFree(tprop.value);
+    }
+
+    g_free(data);
     return retval;
 }
 
 /**
  * Sets the bg if nautilus is appearing to draw the desktop image.
  *
- * TODO: Fix this crap array stuff
- *
  * Simply calls gconftool-2 for now, until we find a better way to do it.
  */
 bool SetBG::set_bg_nautilus(Glib::RefPtr<Gdk::Screen> screen, Glib::ustring file, SetMode mode, Gdk::Color bgcolor) {
 
 	GError *error = NULL;
 
-	Glib::ustring strmode;
+    Glib::ustring strmode = "scaled";       // in case of more modes
 	switch(mode) {
 		case SetBG::SET_SCALE:  strmode = "stretched";  break;
 		case SetBG::SET_TILE:   strmode = "wallpaper"; break; 
 		case SetBG::SET_CENTER: strmode = "centered"; break;
 		case SetBG::SET_ZOOM:   strmode = "scaled"; break;
+		case SetBG::SET_ZOOM_FILL:   strmode = "scaled"; break;
 	};
 
-	gchar** argv = (gchar**)g_malloc(sizeof(gchar*)*10);
-	argv[0] = "gconftool-2";
-	argv[1] = "--type";
-	argv[2] = "string";
-	argv[3] = "--set";
-	argv[4] = "/desktop/gnome/background/picture_options";
-	argv[5] = g_strdup(strmode.c_str());
-	argv[6] = "--set";
-	argv[7] = "/desktop/gnome/background/picture_filename";
-	argv[8] = g_strdup(file.c_str());
-	argv[9] = NULL;
-
-	gboolean res = g_spawn_async(		NULL,
-										argv,
-										NULL,
-										G_SPAWN_SEARCH_PATH,
-										NULL,
-										NULL,
-										NULL,
-										&error);
-
-	if (res == FALSE)
-	{
-		std::cerr << _("ERROR") << "\n" << error->message << "\n";
-		g_clear_error(&error);
+    std::vector<std::string> vecCmdLine;
+    vecCmdLine.push_back(std::string("gconftool-2"));
+    vecCmdLine.push_back(std::string("--type"));
+    vecCmdLine.push_back(std::string("string"));
+    vecCmdLine.push_back(std::string("--set"));
+    vecCmdLine.push_back(std::string("/desktop/gnome/background/picture_options"));
+    vecCmdLine.push_back(std::string(strmode));
+    vecCmdLine.push_back(std::string("--set"));
+    vecCmdLine.push_back(std::string("/desktop/gnome/background/picture_filename"));
+    vecCmdLine.push_back(std::string(file));
+    
+    try
+    {
+        Glib::spawn_async("", vecCmdLine, Glib::SPAWN_SEARCH_PATH);
+    }
+    catch (Glib::SpawnError e)
+    {
+		std::cerr << _("ERROR") << "\n" << e.what() << "\n";
 		
-		for (int i=0; i<9; i++)
-			std::cerr << argv[i] << " ";
+        for (std::vector<std::string>::const_iterator i = vecCmdLine.begin(); i != vecCmdLine.end(); i++)
+			std::cerr << *i << " ";
 
 		std::cerr << "\n";
+
+        return false;
 	}
 
-
-	g_free(argv[5]);
-	g_free(argv[8]);
-	g_free(argv);
-
-	// disregard error atm, just return true or false
-	return (res == TRUE);
+	return true;
 }
 
 /**
@@ -599,6 +634,55 @@ Glib::RefPtr<Gdk::Pixbuf> SetBG::make_zoom(const Glib::RefPtr<Gdk::Pixbuf> orig,
 	return retval;
 }		
 
+/**
+ * Handles SET_ZOOM_FILL mode.
+ *
+ * @param	orig	The original pixbuf
+ * @param	winw	Width of the window
+ * @param	winh	Height of the window
+ * @param	bgcolor	Background color
+ */
+Glib::RefPtr<Gdk::Pixbuf> SetBG::make_zoom_fill(const Glib::RefPtr<Gdk::Pixbuf> orig, const gint winw, const gint winh, Gdk::Color bgcolor) {
+		
+	int x, y, w, h;
+		
+	// depends on bigger side
+	unsigned orig_w = orig->get_width();
+	unsigned orig_h = orig->get_height();
+
+    int dw = winw - orig_w;
+    int dh = winh - orig_h;
+
+    // what if we expand it to fit the screen width?
+    x = 0;
+    w = winw;
+    h = winw * orig_h / orig_w;
+    y = (h - winh) / 2;
+
+    if (!(h >= winh)) {
+        // the image isn't tall enough that way!
+        // expand it to fit the screen height
+        y = 0;
+        w = winh * orig_w / orig_h;
+        h = winh;
+        x = (w - winw) / 2;
+    }
+
+	Glib::RefPtr<Gdk::Pixbuf> tmp = orig->scale_simple(w, h,
+		Gdk::INTERP_BILINEAR);
+	Glib::RefPtr<Gdk::Pixbuf> retval = Gdk::Pixbuf::create(
+		orig->get_colorspace(), orig->get_has_alpha(),
+		orig->get_bits_per_sample(), winw, winh);
+
+	// use passed bg color
+	retval->fill(GdkColorToUint32(bgcolor));
+
+	// copy it in
+	tmp->copy_area(x, y, winw, winh, retval, 0, 0);
+
+	return retval;
+}
+
 /**
  * Utility function to convert a mode (an enum) to a string. 
  */
@@ -619,6 +703,9 @@ Glib::ustring SetBG::mode_to_string( const SetMode mode ) {
 		case SET_ZOOM:
 			ret = Glib::ustring(_("Zoom"));
 			break;
+		case SET_ZOOM_FILL:
+			ret = Glib::ustring(_("ZoomFill"));
+			break;
         case SET_AUTO:
             ret = Glib::ustring(_("Auto"));
             break;
@@ -641,6 +728,8 @@ SetBG::SetMode SetBG::string_to_mode( const Glib::ustring str ) {
 		return SetBG::SET_TILE;
 	else if ( str == Glib::ustring(_("Zoom")) )
 		return SetBG::SET_ZOOM;
+	else if ( str == Glib::ustring(_("ZoomFill")) )
+		return SetBG::SET_ZOOM_FILL;
     else if ( str == Glib::ustring(_("Auto")) )
         return SetBG::SET_AUTO;
 
