@@ -38,14 +38,24 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 class SetBG {
 	public:
 			
+        // NOTE: do not change the order of these, they are changed to integer and 
+        // put on disk.
 		enum SetMode { 
 			SET_SCALE,
 			SET_TILE,
 			SET_CENTER,
 			SET_ZOOM,
-            SET_AUTO
+            SET_AUTO,
+            SET_ZOOM_FILL
 		};
-		
+
+        enum RootWindowType {
+            DEFAULT,
+            NAUTILUS,
+            XFCE,
+            UNKNOWN
+        };
+
 		static bool set_bg(	Glib::ustring &disp,
 							Glib::ustring file,
 							SetMode mode,
@@ -68,7 +78,7 @@ class SetBG {
 									Gdk::Color bgcolor
 								    );
 
-		static bool nautilus_running(Glib::RefPtr<Gdk::Window> rootwin);
+		static SetBG::RootWindowType get_rootwindowtype(Glib::RefPtr<Gdk::Window> rootwin);
 
 		static Glib::ustring mode_to_string( const SetMode mode );
 		static SetMode string_to_mode( const Glib::ustring str );
@@ -79,6 +89,7 @@ class SetBG {
 		static Glib::RefPtr<Gdk::Pixbuf> make_tile(const Glib::RefPtr<Gdk::Pixbuf>, const gint, const gint, Gdk::Color);
 		static Glib::RefPtr<Gdk::Pixbuf> make_center(const Glib::RefPtr<Gdk::Pixbuf>, const gint, const gint, Gdk::Color);
 		static Glib::RefPtr<Gdk::Pixbuf> make_zoom(const Glib::RefPtr<Gdk::Pixbuf>, const gint, const gint, Gdk::Color);
+		static Glib::RefPtr<Gdk::Pixbuf> make_zoom_fill(const Glib::RefPtr<Gdk::Pixbuf>, const gint, const gint, Gdk::Color);
         static SetMode get_real_mode(const Glib::RefPtr<Gdk::Pixbuf>, const gint, const gint);
 
 		static guint32 GdkColorToUint32(const Gdk::Color);
